@@ -16,6 +16,9 @@
 
 #endif
 
+/* deprecated and no longer supported in gcc 3.3 */
+//#define ACCEPT_CR_IN_STRINGS
+
 /* __VA_ARGS__ and __func__ support */
 #define C99_MACROS
 
@@ -1303,6 +1306,16 @@ void bitfield_test(void)
     st1.f5++;
     printf("%d %d %d %d %d\n",
            st1.f1, st1.f2, st1.f3, st1.f4, st1.f5);
+
+    st1.f1 = 7;
+    if (st1.f1 == -1) 
+        printf("st1.f1 == -1\n");
+    else 
+        printf("st1.f1 != -1\n");
+    if (st1.f2 == -1) 
+        printf("st1.f2 == -1\n");
+    else 
+        printf("st1.f2 != -1\n");
 }
 
 #define FTEST(prefix, type, fmt)\
@@ -1589,6 +1602,15 @@ void longlong_test(void)
           printf("%lld\n", value(&a));
     }
     lloptest(0x80000000, 0);
+
+    /* another long long spill test */
+    {
+        long long *p, v;
+        v = 1;
+        p = &v;
+        p[0]++;
+        printf("%lld\n", *p);
+    }
 }
 
 void vprintf1(const char *fmt, ...)
@@ -1661,6 +1683,8 @@ ntf("aaa=%d\n", 3);
     pri\
 \
 ntf("min=%d\n", 4);
+
+#ifdef ACCEPT_CR_IN_STRINGS
     printf("len1=%d\n", strlen("
 "));
 #ifdef CORRECT_CR_HANDLING
@@ -1671,6 +1695,7 @@ ntf("min=%d\n", 4);
     printf("len1=%d\n", strlen("
 a
 "));
+#endif /* ACCEPT_CR_IN_STRINGS */
 }
 
 int reltab[3] = { 1, 2, 3 };
